//-----------------------------------------------------------------------------
// Copyright (c) 2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This application demonstrates how to render a quad using shaders that fills
// up the entire viewport using OpenGL 3.1 or higher or a forward compatible
// OpenGL 3.0 rendering context. The application first clears the viewport to
// a black color and then draws the full screen quad. The full screen quad's
// fragment shader simply fills the full screen quad with a gray color. The
// full screen quad is rendered as a triangle strip with its vertex coordinates
// specified in normalized device coordinates (NDC). Specifying the vertex
// coordinates in this way means view and/or model transformations are not
// needed when rendering the full screen quad since the NDC vertex positions
// will always be mapped to the corners of the current viewport.
//
//-----------------------------------------------------------------------------

#include <windows.h>
#include <stdexcept>
#include "application.h"
#include "shader.h"

using namespace dhpoware;

class Demo : public GL3Application
{
public:
    Demo();
    virtual ~Demo();

protected:
    virtual void init(int argc, char *argv[]);
    virtual void render();
    virtual void shutdown();
    virtual void update(float elapsedTimeSec);

private:
	void initFullScreenQuad();
    void initShaderProgram();
	void renderFullScreenQuad();
    
	GLuint m_vertexArrayObject;
	GLuint m_vertexBufferObject;
	Shader m_shader;
};

//-----------------------------------------------------------------------------
// Demo class implementation.
//-----------------------------------------------------------------------------

Demo::Demo() : GL3Application("OpenGL 3 - Full Screen Quad")
{
}

Demo::~Demo()
{
}

void Demo::init(int argc, char *argv[])
{
    // Load the shader program used to render the quad.
    // The shader program for the quad needs to be loaded before setting up
    // the vertex array and buffer objects because user-defined generic
    // vertex shader variables are being used.

    initShaderProgram();
	initFullScreenQuad();
}

void Demo::initFullScreenQuad()
{
	// Setup the full screen quad geometry as a triangle strip.
	// The vertex positions are in normalized device coordinates (NDC).
	//
	// In NDC:
	//  (-1,  1) is always the top-left corner of the viewport.
	//  (-1, -1) is always the bottom-left corner of the viewport.
	//  ( 1,  1) is always the top-right corner of the viewport.
	//  ( 1, -1) is always the bottom-right corner of the viewport.
	//
	//	v0---v2
	//	|   / |
	//	| /   |
	//	v1---v3

	float quad[] =
	{
		-1.0f,  1.0f,	// v0 - top left corner
		-1.0f, -1.0f,	// v1 - bottom left corner
		 1.0f,  1.0f,	// v2 - top right corner
		 1.0f, -1.0f	// v3 - bottom right corner
	};

	// Create the vertex array object for the full screen quad.
	
	glGenVertexArrays(1, &m_vertexArrayObject);
	glBindVertexArray(m_vertexArrayObject);

	// Create the Vertex Buffer Object for the full screen quad.

	glGenBuffers(1, &m_vertexBufferObject);
	glBindBuffer(GL_ARRAY_BUFFER, m_vertexBufferObject);
	glBufferData(GL_ARRAY_BUFFER, sizeof(quad), quad, GL_STATIC_DRAW);

	// Map the generic vertex attributes used by the full screen quad's shader
	// program to the full screen quad's vertex data stored in the vertex
	// buffer object.

	m_shader.vertexAttribPointer("PositionNDC", 2, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	m_shader.enableVertexAttribArray("PositionNDC");

	glBindVertexArray(0);
}

void Demo::initShaderProgram()
{
	if (!m_shader.loadVertexShader("Content\\Shaders\\FullScreenQuad.vert"))
		throw std::runtime_error(m_shader.lastError());

	if (!m_shader.loadFragmentShader("Content\\Shaders\\FullScreenQuad.frag"))
		throw std::runtime_error(m_shader.lastError());

	if (!m_shader.link())
		throw std::runtime_error(m_shader.lastError());
}

void Demo::render()
{
    glViewport(0, 0, viewportWidth(), viewportHeight());
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	renderFullScreenQuad();
}

void Demo::renderFullScreenQuad()
{
	m_shader.enable();
	glBindVertexArray(m_vertexArrayObject);
	
	// 4 vertices with 2 floats per vertex = 8 elements total.
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 8);

	glBindVertexArray(0);
	m_shader.disable();
}

void Demo::shutdown()
{
	if (m_vertexArrayObject)
	{
		glDeleteVertexArrays(1, &m_vertexArrayObject);
		m_vertexArrayObject = 0;
	}

	if (m_vertexBufferObject)
	{
		glDeleteBuffers(1, &m_vertexBufferObject);
		m_vertexBufferObject = 0;
	}

    glUseProgram(0);
	m_shader.destroy();
}

void Demo::update(float elapsedTimeSec)
{
    if (keyDown(Keyboard::KEY_LALT) || keyDown(Keyboard::KEY_RALT))
    {
        if (keyPressed(Keyboard::KEY_ENTER))
            toggleFullScreen();
    }

    if (keyPressed(Keyboard::KEY_ESCAPE))
        quit();
}

//-----------------------------------------------------------------------------
// WinMain - application entry point.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    Demo demo;

    demo.preferAnisotropicTextureFiltering();
    demo.preferAntiAliasing();

    return demo.run();
}